#!/usr/bin/python
#####################################################################################
# pico_setup1.6_hv4.0.py
# Version 2.3   Relase Date 05.03.2025
# Script to show you all setting up registers pulled from your UPS PIco HV4.0
# -*- coding: utf-8 -*-
# written by ChatGPT o3-mini-high following our detailed instructions
# and re-written (polished) by Ioannis A. Mourtsiadis info@pimodules.com
#####################################################################################

# Set your desired temperature symbol
# C = Celsius, F = Fahrenheit
degrees = "C"

import sys
import time
import datetime
import os
import re     
import getopt
import smbus
import threading
import subprocess
import RPi.GPIO as GPIO

i2c = smbus.SMBus(1)

############### NEW FUNCTIONS (Device 0x69) #################

def fw_version():
    time.sleep(0.1)
    data = i2c.read_word_data(0x69, 0x38)
    return format(data, "04x")
   
def pcb_version():
    time.sleep(0.1)
    data = i2c.read_byte_data(0x69, 0x35)
    return format(data, "02x")
   
def PIco_model():
    time.sleep(0.1)
    data = i2c.read_byte_data(0x69, 0x36)
    return format(data, "02x")
    
def getCPUtemperature():
    time.sleep(0.1)
    res = os.popen('vcgencmd measure_temp').readline()
    cpu_data = res.replace("temp=","").replace("'C\n","")
    cpu_int = int(float(cpu_data))
    return format(cpu_int, "02d")
    
def getCPUtemperaturePIco():
    time.sleep(0.1)
    data = i2c.read_byte_data(0x69, 0x1a)
    return format(data, "02x")
   
def pcb_battery():
    time.sleep(0.1)
    data = i2c.read_byte_data(0x6b, 0x07)
    if data == 0x4c:
        return "0x4c (Battery LiPO 'L')"
    elif data == 0x49: 
        return "0x49 (Battery Li-Ion 'I')"
    elif data == 0x46:
        return "0x46 (Battery LiFePO4 'F')"
    elif data == 0x48:
        return "0x48 (Battery NiMH 'H')"
    elif data == 0x41:
        return "0x41 (Battery SAL 'A')"
    elif data == 0x43:
        return "0x43 (Super Cap (100F)'C')"
    elif data == 0x44:
        return "0x44 (Super Cap HAT 'D')"
    elif data == 0x6c:
        return "0x6c (Battery LiPO+SC 'l')"
    elif data == 0x69:
        return "0x69 (Battery Li-Ion+SC 'i')"
    elif data == 0x66:
        return "0x66 (Battery LiFePO4+SC 'f')"
    elif data == 0x68:
        return "0x68 (Battery NiMH+SC 'h')"
    elif data == 0x61:
        return "0x61 (Battery SAL+SC 'a')"
    else:
        return "ERROR"
  
def default_battery():
    time.sleep(0.1)
    data = i2c.read_byte_data(0x69, 0x37)
    if data == 0x4c:
        return "0x4c (Battery LiPO 'L')"
    elif data == 0x49:
        return "0x49 (Battery Li-Ion 'I')"
    elif data == 0x46:
        return "0x46 (Battery LiFePO4 'F')"
    elif data == 0x48:
        return "0x48 (Battery NiMH 'H')"
    elif data == 0x41:
        return "0x41 (Battery SAL 'A')"
    elif data == 0x43:
        return "0x43 (Super Cap (100F)'C')"
    elif data == 0x44:
        return "0x44 (Super Cap HAT 'D')"
    elif data == 0x6c:
        return "0x6c (Battery LiPO+SC 'l')"
    elif data == 0x69:
        return "0x69 (Battery Li-Ion+SC 'i')"
    elif data == 0x66:
        return "0x66 (Battery LiFePO4+SC 'f')"
    elif data == 0x68:
        return "0x68 (Battery NiMH+SC 'h')"
    elif data == 0x61:
        return "0x61 (Battery SAL+SC 'a')"
    else:
        return "ERROR"

############### EXISTING SYSTEM FUNCTIONS (Device 0x6b) #################

def bat_runtime():
    time.sleep(0.1)
    data = i2c.read_byte_data(0x6b, 0x01)
    if data == 0xff:
        return "Unlimited - Timer Disabled"
    else:
        return data

def picoRTC():
    time.sleep(0.1)
    seconds = i2c.read_byte_data(0x6a, 0x0)
    minutes = i2c.read_byte_data(0x6a, 0x1)
    hours   = i2c.read_byte_data(0x6a, 0x2)
    mday    = i2c.read_byte_data(0x6a, 0x4)
    month   = i2c.read_byte_data(0x6a, 0x5)
    year    = i2c.read_byte_data(0x6a, 0x6)
    return "20" + format(year, "02x") + "-" + format(month, "02x") + "-" + format(mday, "02x") + " " + format(hours, "02x") + ":" + format(minutes, "02x") + ":" + format(seconds, "02x")

def RPiRTC():
    result = subprocess.run(['sudo', 'hwclock', '-r'], capture_output=True, text=True)
    if result.returncode == 0:
        return result.stdout.strip()
    else:
        return "Error: " + result.stderr.strip()

############### REGISTER FUNCTIONS (Device 0x6b) #################

def pico_state():
    time.sleep(0.1)
    return i2c.read_byte_data(0x6b, 0x00) 

def pwr_mode():
    data = i2c.read_byte_data(0x69, 0x00)
    data = data & 0x01
    if data == 1:
        return "Cable POWERED"
    elif data == 0:
        return "BAT/SCAP POWERED"
    else:
        return "ERROR" 

def VRS232_rate():
    time.sleep(0.1)
    data = i2c.read_byte_data(0x6b, 0x02)
    map_str = "0x" + format(data, '02x')
    if (data & 0xf0) == 0x00:
        map_str += " (PICO_OFF  "
    if (data & 0xf0) == 0x10:
        map_str += " (PICO_4800  "
    if (data & 0xf0) == 0x20:
        map_str += " (PICO_9600  "
    if (data & 0xf0) == 0x30:
        map_str += " (PICO_19200  "
    if (data & 0xf0) == 0x40:
        map_str += " (PICO_38400  "
    if (data & 0xf0) == 0x50:
        map_str += " (PICO_57600  "
    if (data & 0xf0) == 0xf0:
        map_str += " (PICO_115200  "
    if (data & 0x0f) == 0x00:
        map_str += "RPI_OFF)"
    if (data & 0x0f) == 0x01:
        map_str += "RPI_4800)"
    if (data & 0x0f) == 0x02:
        map_str += "RPI_9600)"
    if (data & 0x0f) == 0x03:
        map_str += "RPI_19200)"
    if (data & 0x0f) == 0x04:
        map_str += "RPI_38400)"
    if (data & 0x0f) == 0x05:
        map_str += "RPI_57600)"
    if (data & 0x0f) == 0x0f:
        map_str += "RPI_115200)"
    return map_str
   
def serial_monitor():
    time.sleep(0.1)
    data = i2c.read_byte_data(0x6b, 0x2a)
    map_str = "0x" + format(data, '02x')
    if (data & 0xf0) == 0x00:
        map_str += " (PICO_OFF     "
    if (data & 0xf0) == 0xf0:
        map_str += " (PICO_115200  "
    if (data & 0x0f) == 0x00:
        map_str += "RPI_OFF)"
    if (data & 0x0f) == 0x0f:
        map_str += "RPI_115200)"
    return map_str 

def rpi_hold_on():
    time.sleep(0.1)
    return i2c.read_byte_data(0x6b, 0x03)

def pico_rpi_route():
    time.sleep(0.1)
    return i2c.read_byte_data(0x6b, 0x04)

def sta_timer():
    time.sleep(0.1)
    return i2c.read_byte_data(0x6b, 0x05)

def enable5V():
    time.sleep(0.1)
    return i2c.read_byte_data(0x6b, 0x06)

def setA_D():
    time.sleep(0.1)
    return i2c.read_byte_data(0x6b, 0x08)

def led_green():
    time.sleep(0.1)
    return i2c.read_byte_data(0x6b, 0x09)

def led_blue():
    time.sleep(0.1)
    return i2c.read_byte_data(0x6b, 0x0a)

def relay():
    time.sleep(0.1)
    return i2c.read_byte_data(0x6b, 0x0b)

def bmode():
    time.sleep(0.1)
    return i2c.read_byte_data(0x6b, 0x0c) 

def bmode_monitor():    
    time.sleep(0.1)
    data = i2c.read_byte_data(0x6b, 0x0c)
    map_str = "0x" + format(data, '02x')
    if (data & 0x03) == 0x00:
        map_str += " (BUZZER OFF)"
    if (data & 0x03) == 0x01:
        map_str += " (BUZZER AUTO)"
    if (data & 0x03) == 0x02:
        map_str += " (BUZZER USER)"
    return map_str    
    
def bfreq():
    time.sleep(0.1)
    return i2c.read_word_data(0x6b, 0x0e)

def bdur():
    time.sleep(0.1)
    return i2c.read_byte_data(0x6b, 0x10)

def fmode():
    time.sleep(0.1)
    return i2c.read_byte_data(0x6b, 0x11)
    
def fmode_monitor():
    time.sleep(0.1)
    data = i2c.read_byte_data(0x6b, 0x11)
    map_str = "0x" + format(data, '02x')
    if (data & 0x03) == 0x00:
        map_str += " (FAN OFF)"
    elif (data & 0x03) == 0x01:
        map_str += " (FAN ON)"
    elif (data & 0x03) == 0x02:
        map_str += " (FAN AUTO MILD)"
    elif (data & 0x03) == 0x03:
        map_str += " (FAN AUTO HARD)"
    return map_str


def fspeed():
    time.sleep(0.1)
    return i2c.read_byte_data(0x6b, 0x12)

def fttemp():
    time.sleep(0.1)
    return i2c.read_byte_data(0x6b, 0x13)

def fsstep():
    time.sleep(0.1)
    return i2c.read_byte_data(0x6b, 0x14)

def led_off():
    time.sleep(0.1)
    return i2c.read_byte_data(0x6b, 0x15)

def on_the_go():
    time.sleep(0.1)
    return i2c.read_byte_data(0x6b, 0x16)

def map_led_green():
    time.sleep(0.1)
    return i2c.read_byte_data(0x6b, 0x17)

def map_led_blue():
    time.sleep(0.1)
    return i2c.read_byte_data(0x6b, 0x18)

def chg_ctrl():
    time.sleep(0.1)
    return i2c.read_byte_data(0x6b, 0x19)

def fssd_duration():
    time.sleep(0.1)
    return i2c.read_byte_data(0x6b, 0x1a)

def rpi_OFF_time_reg():
    time.sleep(0.1)
    return i2c.read_byte_data(0x6b, 0x1b)

def ad_corectionf():
    time.sleep(0.1)
    data = i2c.read_byte_data(0x6b, 0x1c)
    if data > 127:
        data -= 256
    return data

def scheduler_sel():
    time.sleep(0.1)
    return i2c.read_byte_data(0x6b, 0x1d)

def bs_action_time():
    time.sleep(0.1)
    return i2c.read_byte_data(0x6b, 0x1e)

def bs_inaction_time():
    time.sleep(0.1)
    return i2c.read_byte_data(0x6b, 0x1f)

def bs_start_time():
    time.sleep(0.1)
    return i2c.read_word_data(0x6b, 0x20)

def bs_multipier():
    time.sleep(0.1)
    return i2c.read_byte_data(0x6b, 0x22)

def bs_run():
    time.sleep(0.1)
    return i2c.read_byte_data(0x6b, 0x23)

def sas_number():
    time.sleep(0.1)
    return i2c.read_byte_data(0x6b, 0x24)

def sas_run():
    time.sleep(0.1)
    return i2c.read_byte_data(0x6b, 0x25)

def ps_checking_time():
    time.sleep(0.1)
    return i2c.read_byte_data(0x6b, 0x26)

def bat_level_report():
    time.sleep(0.1)
    return i2c.read_byte_data(0x6b, 0x27)

def sc_level_start():
    time.sleep(0.1)
    return i2c.read_word_data(0x6b, 0x28)

def debug_mode():
    time.sleep(0.1)
    return i2c.read_byte_data(0x6b, 0x2a)

def rs232_data():
    time.sleep(0.1)
    return i2c.read_i2c_block_data(0x6b, 0x2b, 16)

############### PRINTING REGISTER VALUES ###############

def format_reg_line(label, value, label_field_width=30, value_field_width=35):
    if label in reg_address_map:
        addr_str = f" ({reg_address_map[label]})"
    else:
        addr_str = ""
    dot_count = max(0, (label_field_width - len(label) - len(addr_str)) - 6)
    dots = '.' * dot_count
    left_part = f"{label}{dots}{addr_str} :"
    right_part = f" {value:<{value_field_width}}"
    return f"{left_part}{right_part}"

reg_address_map = {
    "pico_state": "0x00",
    "Battery Runtime": "0x01",
    "RS232 Rate": "0x02",
    "rpi_hold_on": "0x03",
    "pico_rpi_route": "0x04",
    "sta_timer": "0x05",
    "enable5V": "0x06",
    "battype": "0x07",
    "setA_D": "0x08",
    "led_green": "0x09",
    "led_blue": "0x0a",
    "relay": "0x0b",
    "bmode": "0x0c",
    "bfreq": "0x0e",
    "bdur": "0x10",
    "fmode": "0x11",
    "fspeed": "0x12",
    "fttemp": "0x13",
    "fsstep": "0x14",
    "LEDs_ON/OFF": "0x15",
    "on_the_go": "0x16",
    "map_led_green": "0x17",
    "map_led_blue": "0x18",
    "chg_ctrl": "0x19",
    "fssd_duration": "0x1a",
    "rpi_OFF_time_reg": "0x1b",
    "ad_corectionf": "0x1c",
    "Scheduler_Sel": "0x1d",
    "BS_action_time": "0x1e",
    "BS_inaction_time": "0x1f",
    "BS_Start_time": "0x20",
    "BS_multipier": "0x22",
    "BS_RUN": "0x23",
    "SAS_number": "0x24",
    "SAS_run": "0x25",
    "ps_checking_time": "0x26",
    "BAT_level_report": "0x27",
    "SC_level_start": "0x28",
    "debug_mode": "0x2a",
    "RS232_data": "0x2b"
}

############### REGISTER LISTING FUNCTIONS ###############
def get_all_registers():
    runtime = bat_runtime()
    if isinstance(runtime, int):
        runtime_str = format(runtime, '#04x') + " (" + str(runtime) + " min)"
    else:
        runtime_str = runtime
    en5V = enable5V()
    if en5V == 0:
        enable5V_str = format(en5V, '#04x') + " (OFF)"
    elif en5V == 1:
        enable5V_str = format(en5V, '#04x') + " (ON)"
    else:
        enable5V_str = format(en5V, '#04x')
    ledoff = led_off()
    if ledoff == 0:
        led_off_str = format(ledoff, '#04x') + " (OFF)"
    elif ledoff == 1:
        led_off_str = format(ledoff, '#04x') + " (ON)"
    else:
        led_off_str = format(ledoff, '#04x')
    led_green_val = led_green()
    if led_green_val == 0:
        led_green_str = format(led_green_val, '#04x') + " (OFF)"
    elif led_green_val == 1:
        led_green_str = format(led_green_val, '#04x') + " (ON)"
    else:
        led_green_str = format(led_green_val, '#04x')
    led_blue_val = led_blue()
    if led_blue_val == 0:
        led_blue_str = format(led_blue_val, '#04x') + " (OFF)"
    elif led_blue_val == 1:
        led_blue_str = format(led_blue_val, '#04x') + " (ON)"
    else:
        led_blue_str = format(led_blue_val, '#04x')
    sta_val = sta_timer()
    if isinstance(sta_val, int):
        if sta_val == 0xff:
            sta_timer_str = format(sta_val, '#04x') + " (OFF)"
        else:
            sta_timer_str = format(sta_val, '#04x') + " (" + str(sta_val) + " seconds)"
    else:
        sta_timer_str = sta_val
    relay_val = relay()
    if relay_val == 0:
        relay_str = format(relay_val, '#04x') + " (OFF)"
    elif relay_val == 1:
        relay_str = format(relay_val, '#04x') + " (ON)"
    else:
        relay_str = format(relay_val, '#04x')
    bfreq_val = bfreq()
    bfreq_str = format(bfreq_val, '#06x') + " (" + str(bfreq_val) + " Hz)"
    fspeed_val = fspeed()
    fspeed_str = format(fspeed_val, '#04x') + " (" + str(fspeed_val) + ")"

    items = [
        ("pico_state", format(pico_state(), '#04x')),
        ("Battery Runtime", runtime_str),
        ("RS232 Rate", VRS232_rate()),
        ("rpi_hold_on", format(rpi_hold_on(), '#04x')),
        ("pico_rpi_route", format(pico_rpi_route(), '#04x')),
        ("sta_timer", sta_timer_str),
        ("enable5V", enable5V_str),
        ("battype", pcb_battery()),
        ("setA_D", format(setA_D(), '#04x')),
        ("led_green", led_green_str),
        ("led_blue", led_blue_str),
        ("relay", relay_str),
        ("bmode", bmode_monitor()),
        ("bfreq", bfreq_str),
        ("bdur", format(bdur(), '#04x')),
        ("fmode", fmode_monitor()),  # Ensure fmode is printed using its monitor function
        ("fspeed", fspeed_str),
        ("fttemp", format(fttemp(), '#04x')),
        ("fsstep", format(fsstep(), '#04x')),
        ("LEDs_ON/OFF", led_off_str),
        ("on_the_go", format(on_the_go(), '#04x')),
        ("map_led_green", format(map_led_green(), '#04x')),
        ("map_led_blue", format(map_led_blue(), '#04x')),
        ("chg_ctrl", format(chg_ctrl(), '#04x')),
        ("fssd_duration", format(fssd_duration(), '#04x')),
        ("rpi_OFF_time_reg", format(rpi_OFF_time_reg(), '#04x')),
        ("ad_corectionf", format(ad_corectionf(), '#04x')),
        ("Scheduler_Sel", format(scheduler_sel(), '#04x')),
        ("BS_action_time", format(bs_action_time(), '#04x')),
        ("BS_inaction_time", format(bs_inaction_time(), '#04x')),
        ("BS_Start_time", format(bs_start_time(), '#06x')),
        ("BS_multipier", format(bs_multipier(), '#04x')),
        ("BS_RUN", format(bs_run(), '#04x')),
        ("SAS_number", format(sas_number(), '#04x')),
        ("SAS_run", format(sas_run(), '#04x')),
        ("ps_checking_time", format(ps_checking_time(), '#04x')),
        ("BAT_level_report", format(bat_level_report(), '#04x')),
        ("SC_level_start", format(sc_level_start(), '#06x')),
        ("debug_mode", serial_monitor()),
        ("RS232_data", " ".join(f"0x{b:02x}" for b in rs232_data()[:8]))
    ]

    reg_types = {}
    for name, addr, size, t in get_mod_regs_list():
        reg_types[name.lower()] = t

    mod_items = []
    for reg_name, val in items:
        if reg_types.get(reg_name.lower()) == "b" and " (" in val:
            val = val.replace(" (", "   (", 1)
        mod_items.append((reg_name, val))

    new_items = []
    mod_regs = get_mod_regs_list()
    for i, (label, val) in enumerate(mod_items):
        base_name = label
        type_str = ""
        for (name, addr, size, t) in mod_regs:
            if name.lower() == base_name.lower():
                type_str = t
                break
        new_label = f"{i}({type_str}) " + label
        new_items.append((new_label, val))

    return new_items  # Ensure this return is reached


def new_print_registers_two_columns():
    # Print the header above the table exactly as provided
    print("+----------------------------------------------------------------------------------------------------------------------+")
    print("|Number|     Register    |Address|Value|                                        |Number|     Register    |Address|Value|")
    print("+----------------------------------------------------------------------------------------------------------------------+")
    
    items = get_all_registers()
    total = len(items)
    half = (total + 1) // 2
    separator = " " * 16 + "| "

    for i in range(half):
        # Use the label as returned (which already includes the index and type)
        left_label = f"{items[i][0]:.<22}"
        # Extract the register name from the label (assumes the label is in the form "i(type) register_name")
        parts_left = items[i][0].split(maxsplit=1)
        reg_name_left = parts_left[1] if len(parts_left) > 1 else items[i][0]
        left_addr = reg_address_map.get(reg_name_left, "0x00")
        left_value = f"{'.'*4}{left_addr} : {items[i][1]}"
        if reg_name_left == "RS232_data":
            data_values = items[i][1].split() if isinstance(items[i][1], str) else items[i][1]
            limited_values = " ".join(data_values[:8])
            left_value = f"{'.'*4}{left_addr} : {limited_values}"
        left_part = f"{left_label}{left_value:<40}"

        right_part = ""
        if i + half < total:
            right_label = f"{items[i+half][0]:.<22}"
            parts_right = items[i+half][0].split(maxsplit=1)
            reg_name_right = parts_right[1] if len(parts_right) > 1 else items[i+half][0]
            right_addr = reg_address_map.get(reg_name_right, "0x00")
            right_value = f"{'.'*4}{right_addr} : {items[i+half][1]}"
            if reg_name_right == "RS232_data":
                data_values = items[i+half][1].split() if isinstance(items[i+half][1], str) else items[i+half][1]
                limited_values = " ".join(data_values[:8])
                right_value = f"{'.'*4}{right_addr} : {limited_values}"
            right_part = f"{right_label}{right_value:<40}"

        print(f"  {left_part}{separator}{right_part}")


############### UPDATE HELPER (Modifiable Registers) ###############

modifiable_registers = {
    "Pico_State":           (0x00, 1),
    "Battery Runtime":      (0x01, 1),
    "RS232 Rate":           (0x02, 1), 
    "rpi_hold_on":          (0x03, 1),
    "pico_rpi_route":       (0x04, 1),
    "sta_timer":            (0x05, 1),
    "enable5V":             (0x06, 1),
    "battype":              (0x07, 1),
    "setA_D":               (0x08, 1),
    "led_green":            (0x09, 1),
    "led_blue":             (0x0A, 1), 
    "relay":                (0x0B, 1),
    "bmode":                (0x0C, 1), 
    "bfreq":                (0x0E, 2), 
    "bdur":                 (0x10, 1),
    "fmode":                (0x11, 1), 
    "fspeed":               (0x12, 1), 
    "fttemp":               (0x13, 1),
    "fsstep":               (0x14, 1), 
    "LEDs_ON/OFF":          (0x15, 1),
    "on_the_go":            (0x16, 1), 
    "map_led_green":        (0x17, 1),
    "map_led_blue":         (0x18, 1),
    "chg_ctrl":             (0x19, 1),
    "fssd_duration":        (0x1A, 1),
    "rpi_OFF_time_reg":     (0x1B, 1),
    "ad_corectionf":        (0x1C, 1),
    "Scheduler_Sel":        (0x1D, 1),
    "BS_action_time":       (0x1E, 1),
    "BS_inaction_time":     (0x1F, 1),
    "BS_Start_time":        (0x20, 2),
    "BS_multipier":         (0x22, 1),
    "BS_RUN":               (0x23, 1),
    "SAS_number":           (0x24, 1),
    "SAS_run":              (0x25, 1),
    "ps_checking_time":     (0x26, 1),
    "BAT_level_report":     (0x27, 1),
    "SC_level_start":       (0x28, 2),
    "debug_mode":           (0x2A, 1),
    "RS232_data":           (0x2B, 8)
}

def get_mod_regs_list():
    regs = []
    for name, value in modifiable_registers.items():
        if isinstance(value, tuple):
            addr, size = value
        else:
            addr, size = value, 1
        if size == 1:
            type_str = "b"
        elif size == 2:
            type_str = "w"
        else:
            type_str = "blk"
        regs.append((name, addr, size, type_str))
    return regs

def update_register(reg_name, reg_addr, size):
    try:
        if size == 1:
            current_val = i2c.read_byte_data(0x6b, reg_addr)
        elif size == 2:
            current_val = i2c.read_word_data(0x6b, reg_addr)
        else:
            current_val = None
    except Exception as e:
        print(f"Error reading {reg_name} (Register 0x{reg_addr:02x}): {e}")
        return
    print("\n{:<20} | Current value: 0x{:x}".format(reg_name, current_val))
    if size == 1:
        prompt = "Enter new value for byte register (e.g. 255 or 0xFF), or press ENTER to keep unchanged: "
        max_val = 255
    elif size == 2:
        prompt = "Enter new value for word register (e.g. 65535 or 0xFFFF), or press ENTER to keep unchanged: "
        max_val = 65535
    else:
        prompt = "Enter new value, or press ENTER to keep unchanged: "
        max_val = 255

    if reg_name == "Pico_State":
        allowed_values = [0x11, 0x22, 0x33, 0x99, 0xaa, 0xbb, 0xcc, 0xdd, 0xee, 0xff]
        prompt = "Enter new value for Pico_State register (Allowed: " + ", ".join(format(v, '#04x') for v in allowed_values) + "): "
    elif reg_name == "enable5V":
        allowed_values = [0x00, 0x01]
        prompt = "Enter new value for enable5V register (Allowed: " + ", ".join(format(v, '#04x') for v in allowed_values) + "): "
    elif reg_name == "battype":
        allowed_values = [0x4c, 0x49, 0x46, 0x48, 0x41, 0x43, 0x44, 0x6c, 0x69, 0x66, 0x68, 0x61]
        prompt = "Enter new value for battype register (Allowed: " + ", ".join(format(v, '#04x') for v in allowed_values) + "): "
    elif reg_name == "on_the_go":
        allowed_values = [0xaa]
        prompt = "Enter new value for on_the_go register (Allowed: " + ", ".join(format(v, '#04x') for v in allowed_values) + "): "
    elif reg_name == "fmode":
        allowed_values = [0x00, 0x01, 0x02, 0x03, 0xff]
        prompt = "Enter new value for fmode register (Allowed: " + ", ".join(format(v, '#04x') for v in allowed_values) + "): "
    elif reg_name == "RS232 Rate":
        # Allowed combinations are any OR of:
        # Lower nibble: 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x0f
        # Upper nibble: 0x00, 0x10, 0x20, 0x30, 0x40, 0x50, 0xf0
        prompt = ("Enter new value for RS232 Rate register (Allowed OR combinations of:\n"
                  "  Lower nibble: 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x0f\n"
                  "  Upper nibble: 0x00, 0x10, 0x20, 0x30, 0x40, 0x50, 0xf0\n"
                  "For example: 0x11 (0x01|0x10), 0x1f (0x0f|0x10), etc.): ")

    new_val_str = input(prompt).strip()
    if new_val_str == "":
        print(f"{reg_name} remains unchanged.")
        return
    try:
        if new_val_str.lower().startswith("0x"):
            new_val = int(new_val_str, 16)
        else:
            new_val = int(new_val_str)
        if not (0 <= new_val <= max_val):
            print(f"Value must be between 0 and {max_val}. Skipping update for this register.")
            return
        if reg_name in ["Pico_State", "enable5V", "battype", "on_the_go", "fmode"]:
            if new_val not in allowed_values:
                print("Invalid value for " + reg_name + " register. Allowed values are: " + ", ".join(format(v, '#04x') for v in allowed_values))
                return
        elif reg_name == "RS232 Rate":
            lower_allowed = {0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x0f}
            upper_allowed = {0x00, 0x10, 0x20, 0x30, 0x40, 0x50, 0xf0}
            lower_nibble = new_val & 0x0f
            upper_nibble = new_val & 0xf0
            if lower_nibble not in lower_allowed or upper_nibble not in upper_allowed:
                print("Invalid value for RS232 Rate register. Allowed combinations are any OR combination of the following:")
                print("  Lower nibble: 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x0f")
                print("  Upper nibble: 0x00, 0x10, 0x20, 0x30, 0x40, 0x50, 0xf0")
                return
    except ValueError:
        print("Invalid input. Skipping update for this register.")
        return
 
    try:
        if size == 1:
            i2c.write_byte_data(0x6b, reg_addr, new_val)
        elif size == 2:
            i2c.write_word_data(0x6b, reg_addr, new_val)
        time.sleep(0.1)
        print(f"{reg_name} updated to 0x{new_val:x}.")
    except Exception as e:
        print(f"Error writing to {reg_name} (Register 0x{reg_addr:02x}): {e}")


############### DISPLAY SYSTEM STATUS ###############

def display_status():
    print("\n*************************************************")
    print("           UPS PIco HV4.0 0x6B Setup")
    print("      Version 2.3   Relase Date 05.03.2025")
    print("     Compatible with Firmware 0x0180 and up")
    print("*************************************************")
    print(" Powered by PiModules with help from ChatGPT 4.0 ")
    print("*************************************************\n")
    os.system("sudo hwclock --systohc")
    time.sleep(0.5)
    print("{:<30} : {}".format("System Clock (HWCLOCK)", RPiRTC()))
    print("{:<30} : {}".format("UPS PIco RTC Registers", picoRTC()))
    print("")
    print("{:<30} : {}".format("UPS PIco Firmware", fw_version()))
    fw_int = int(fw_version(), 16)
    if fw_int < int("0x0180", 16):
        print("Wrong Firmware Version - Please update it to 0x0180 and up")
        print("")
        print("")
        exit()
    print("{:<30} : {}".format("UPS PIco Bootloader", pcb_version()))
    print("{:<30} : {}".format("UPS PIco PCB Version", PIco_model()))
    print("")
    print("{:<30} : {}".format("UPS PIco PCB Defaut Battery", pcb_battery()))
    print("{:<30} : {}".format("UPS PIco Set Battery", default_battery()))
    print("{:<30} : {}".format("Powering Mode", pwr_mode()))
    print("")
    print("{:<30} : {} C".format("RP Core Temp", int(getCPUtemperature(), 10)))
    print("{:<30} : {} C".format("RP Core Temp (PICo)", int(getCPUtemperaturePIco(), 10)))
    print("")

############### MAIN SCRIPT ###############

def main():
    display_status()
    new_print_registers_two_columns()
    
    mod_regs_list = get_mod_regs_list()
    while True:
        choice = input("\nEnter the number of the register you want to modify (or press ENTER to finish): ").strip()
        if choice == "":
            break
        try:
            num = int(choice)
        except ValueError:
            print("Invalid input. Please enter a number.")
            continue
        if num < 0 or num >= len(mod_regs_list):
            print("Invalid register number. Try again.")
            continue
        reg_name, reg_addr, reg_size, type_str = mod_regs_list[num]
        update_register(reg_name, reg_addr, reg_size)
    
    print("\n--- Updated Register Listing ---\n")
    new_print_registers_two_columns()

if __name__ == "__main__":
    main()
